/*
 * A converter from cm to cc files
 *
 * This is basically a preprocessor for almost-C++ files. It is used as
 *   cm <file.cm >file.cc
 * and makes only some small modifications to the .cm source:
 *
 * 1. Every division a/b is translated into divconv(a)/b. This allows for a
 *    a statement
 *      rational r = 1/2;
 *    which in plain C++ will always be treated as integer division and
 *    therefore give 0, which is not intended. But after translation with cm,
 *    this becomes
 *      rational r = divconv(1)/2;
 *    which is now interpreted correctly, if divconv is defined to be a
 *    typecast to a type that allows fractions.
 * 2. To still allow ordinary integer division, "/." is translated into "/".
 * 3. Exponents: For some weird reasons, "^" is the xor operator in C++, which
 *    has a very low precedence (so just overloading "^" to an exponential
 *    function does not work). We replace a^b by pow(a,b) with the correct
 *    precedences, i.e.
 *      a^2+b^2
 *    in the .cm file will give
 *      pow(a,2)+pow(b,2)
 *    in the .cc file.
 * 4. A GW function is one that takes an argument of type gwlist. The GW
 *    functions are listed in expr.h . We replace
 *      GW (x,y,z,...)
 *    by
 *      GW (gwlist(x)+gwlist(y)+gwlist(z)+...)
 *    where every x,y,z,... can be of the special form
 *      x  or  [x,m]  or  x:c  or  [x,m]:c
 *    which is translated into gwlist(x), gwlist (x,m), gwlist (x,0,c),
 *    gwlist (x,m,c), respectively.
 * 5. We include standard header files: iostream.h, gwinv.h.
 *
 * Other than that, no translations are made. Even whitespaces and linebreaks
 * are copied without change.
 *
 * Note that the above replacements are (hopefully) smart; we parse the
 * source code in a rudimentary way to set the brackets at the right
 * position. String constants and comments are understood and remain
 * unchanged.
 *
 */

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "inout.h"
#include "expr.h"



int main (int argc, char *argv[])
{
	char *s;

	s = inout_init ();

	write_string ("// *** This file was automatically "
		"generated by cm ***\n\n");
	write_string ("#include <iostream.h>\n");
	write_string ("#include \"gwinv.h\"\n\n");
	write_flush ();

	for (;;) {
		// See if there is something left to read
		s = write_white ();
		if ((*s) == ENOF) break; // end of file

		// There is still something to read. Do so.
		read_expr (s,1);
		write_flush ();
	}

	exit (0);
}

