/*
 * expr.h - parse the input line and read an expression
 *
 */

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "expr.h"



typedef enum {END,BINARY,PREFIX,POSTFIX,RESERVED,GWTYPE,GWKEY} optype_t;

class op_t {
	public:
		optype_t type;
		int lvl; // The precedence level
		char *name;

		op_t (optype_t t, int n, char *s) {
			type = t;lvl = n; name = s;
		};
};



// Every operator (except RESERVED and GWKEY) has a precedence level.
// Currently, we're only changing '/', '^', '/.', '^.', so these operators
// get levels for their own. In between, it doesn't matter if we assign same
// levels to operators that will actually have different precedences in C++,
// as long as their relative position is OK to those that we change.
//
// So:
// - level 1 are the separators appearing in the GW functions, currently only
//   "," and ":"
// - level 2 is everything below division.
// - level 3 is '/' and '/.'
// - level 4 is everything between division and power
// - level 5 is '^'
// - level 6 is everything above '^'.
//
// The reserved words are those that do not take any arguments before or
// after them. It is important do distinguish them from e.g. variable names,
// because return (6)^3; has to be translated to return pow((6),3); , whereas
// xy(6)^3 has to be translated to pow(xy(6),3); .
// This is special only to those C++ keywords that take numbers as arguments
// that don't need brackets around them.

#define MAXLEVEL 6
#define EXPLEVEL 5
#define DIVLEVEL 3

// List of operators. They should be ordered from longest to shortest, so that
// an early match doesn't catch a partial longer operator.

const op_t op[] = {
	op_t (GWKEY,100,"rational"),
	op_t (GWKEY,101,"elliptic"),
	op_t (GWKEY,200,"points"),
	op_t (GWKEY,201,"lines"),
	op_t (GWKEY,202,"conics"),
	op_t (GWKEY,202,"quadrics"),
	op_t (GWKEY,203,"cubics"),
	op_t (GWKEY,204,"quartics"),
	op_t (GWKEY,205,"quintics"),
	op_t (GWKEY,206,"sextics"),
	op_t (GWKEY,207,"septics"),
	op_t (GWKEY,208,"octics"),
	op_t (GWKEY,209,"nonics"),
	op_t (GWKEY,302,"quadric"),
	op_t (GWKEY,303,"cubic"),
	op_t (GWKEY,304,"quartic"),
	op_t (GWKEY,305,"quintic"),
	op_t (GWKEY,306,"sextic"),
	op_t (GWKEY,307,"septic"),
	op_t (GWKEY,308,"octic"),
	op_t (GWKEY,309,"nonic"),
	op_t (GWKEY,400,"point"),
	op_t (GWKEY,401,"curve"),
	op_t (GWKEY,402,"surface"),
	op_t (GWKEY,403,"threefold"),
	op_t (GWKEY,403,"3fold"),
	op_t (GWKEY,404,"fourfold"),
	op_t (GWKEY,404,"4fold"),
	op_t (GWKEY,405,"5fold"),
	op_t (GWKEY,406,"6fold"),
	op_t (GWKEY,407,"7fold"),
	op_t (GWKEY,408,"8fold"),
	op_t (GWKEY,409,"9fold"),
	op_t (GWKEY,500,"on"),
	op_t (GWKEY,500,"in"),
	op_t (GWKEY,501,"to"),
	op_t (GWKEY,502,"through"),
	op_t (GWKEY,502,"thru"),
	op_t (GWKEY,503,"with"),
	op_t (GWKEY,504,"multiplicities"),
	op_t (GWKEY,504,"multiplicity"),
	op_t (GWKEY,504,"mults"),
	op_t (GWKEY,504,"mult"),
	op_t (RESERVED,0,"return"),
	op_t (RESERVED,0,"new"),
	op_t (RESERVED,0,"delete"),
	op_t (RESERVED,0,"sizeof"),
	op_t (GWTYPE,0,"GW"),
	op_t (BINARY,2,"<<="),
	op_t (BINARY,2,">>="),
	op_t (BINARY,4,"->*"),
	op_t (BINARY,2,"+="),
	op_t (BINARY,2,"-="),
	op_t (BINARY,2,"*="),
	op_t (BINARY,2,"/="),
	op_t (BINARY,2,"%="),
	op_t (BINARY,2,"&="),
	op_t (BINARY,2,"|="),
	op_t (BINARY,2,"^="),
	op_t (BINARY,2,"||"),
	op_t (BINARY,2,"&&"),
	op_t (BINARY,2,"=="),
	op_t (BINARY,2,"!="),
	op_t (BINARY,2,"<="),
	op_t (BINARY,2,">="),
	op_t (BINARY,2,"<<"),
	op_t (BINARY,2,">>"),
	op_t (PREFIX,4,"++"),
	op_t (PREFIX,4,"--"),
	op_t (POSTFIX,4,"++"),
	op_t (POSTFIX,4,"--"),
	op_t (BINARY,6,"->"),
	op_t (BINARY,4,".*"),
	op_t (BINARY,6,"::"),
	op_t (PREFIX,6,"::"),
	op_t (BINARY,3,"/."),
	op_t (RESERVED,0,";"),
	op_t (POSTFIX,2,";"),
	op_t (BINARY,1,","),
	op_t (BINARY,2,"="),
	op_t (BINARY,2,"?"),
	op_t (BINARY,1,":"),
	op_t (BINARY,2,"|"),
	op_t (BINARY,2,"&"),
	op_t (BINARY,2,"<"),
	op_t (BINARY,2,">"),
	op_t (BINARY,2,"+"),
	op_t (BINARY,2,"-"),
	op_t (BINARY,2,"*"),
	op_t (PREFIX,4,"*"),
	op_t (BINARY,2,"%"),
	op_t (BINARY,3,"/"),
	op_t (PREFIX,4,"*"),
	op_t (PREFIX,4,"&"),
	op_t (PREFIX,4,"+"),
	op_t (PREFIX,4,"-"),
	op_t (PREFIX,4,"!"),
	op_t (PREFIX,4,"~"),
	op_t (BINARY,5,"^"),
	op_t (BINARY,6,"."),
	op_t (END,0,"")
};



bool is_alnum (char c) {
	return (c>='0' && c<='9') || (c=='_') ||
		    (c>='A' && c<='Z') || (c>='a' && c<='z');
}



// Check if there is an operator of type 'type' at *s. Return its precedence
// level, or -1 if no operator was found. Set num to the number of characters
// read. Nothing is written to the output queue, and the input pointer is
// not advanced.

int read_op (char *s, int& num, optype_t type)
{
	int i = -1;
	int j;
	char c;

	num = 0;

	while (op[++i].type != END) if (op[i].type == type) {
		// Check if we can find operator i
		j = 0;
		while (op[i].name[j] == s[num+j]) {
			j++;
			if (!op[i].name[j]) goto FOUND;
			if (s[num+j]<' ' || s[num+j]==ENOF) break; // not found
		}
	}

	// No operator found.
	return -1;

FOUND:	// The operator was found.
	num += j;
	// If the operator was a word, a non-alphabetic character has to
	// follow.
	if ( is_alnum (op[i].name[j-1]) && is_alnum (s[num]) ) return -1;

	return op[i].lvl;
}



int read_term (char *s, int& num) // Same as read_op, but reads a name, a
				  // number, a string or a char constant.
				  // Returns -1 if nothing was found, 0
				  // otherwise.
{
	num = 0;

	// String constant
	if (s[num]=='"') {
		while (s[++num] != '"') {
			if (s[num]<' ' || s[num]==ENOF) return -1;
			if (s[num]=='\\') num++;
		}
		num++;
		return 0;
	}

	// Char constant
	if (s[num]=='\'') {
		while (s[++num] != '\'') {
			if (s[num]<' ' || s[num]==ENOF) return -1;
			if (s[num]=='\\') num++;
		}
		num++;
		return 0;
	}

	// Name or number
	if ( is_alnum (s[num]) ) {
		while ( is_alnum (s[++num]) );
		return 0;
	}

	return -1;
}



char read_open_bracket (char *s, int& num) // Same as read_op, but reads an
	// opening bracket (, [ or {. Returns this character if found,
	// otherwise 0.
{
	num=1;

	if ((*s)=='(' || (*s)=='[' || (*s)=='{') return *s;
	else return 0;
}



char read_close_bracket (char *s, char c, int& num) // Reads the closing
	// bracket to c.
{
	num=1;

	if ((*s)==')' && c=='(') return *s;
	if ((*s)==']' && c=='[') return *s;
	if ((*s)=='}' && c=='{') return *s;

	return 0;
}



void write_stuff (char *&s, int num)
{
	for (int i=0; i<num; i++) {
		s = write_next ();
	}
	s = write_white ();
}



void read_gwarg (char *& s) // Read an argument for the GW function
{
	int m=0; // do we have a multiplicity?
	int i,k,n;

START:	// Do we have a keyword?
	k = read_op (s,n,GWKEY);
	if (k>=0 || (*s)=='P') {
		// Yes...
		if ((*s)=='P') {
			// The special keyword Pn (projective space).
			s = skip_next ();
			s = write_white ();
			if ( (*s)=='^' ) {
				s = skip_next ();
				s = write_white ();
			}
			write_string ("E=1)+gwlist(N=");
			read_expr (s,2);
			s = write_white ();
			goto AGAIN;
		}	
		// We have an actual keyword.
		for (i=1; i<=n; i++) s = skip_next ();
		s = write_white ();
		if (k>=100 && k<110) {
			write_string ("G=");
			write_char (k-100+'0');
			goto AGAIN;
		}
		if (k>=200 && k<210) {
			write_string ("D=");
			write_char (k-200+'0');
			goto AGAIN;
		}
		if (k>=300 && k<310) {
			write_string ("E=");
			write_char (k-300+'0');
			goto AGAIN;
		}
		if (k>=400 && k<410) {
			write_string ("N=");
			write_char (k-400+'0');
			goto AGAIN;
		}
		if (k>=500) goto START;
		else stop_error ("invalid keyword");
	AGAIN:	// After a keyword, we don't need a comma
		if ( (*s)==',' ) return;
		if ( (*s)==')' ) return;
		write_string (")+gwlist(");
		goto START;
	}

	// Read '[' if present
	if ( (*s)=='[' ) {
		s = skip_next ();
		s = write_white ();
		m = 1;
	}

	// Read the class, and write it to the output
	read_expr (s,2);
	s = write_white ();

	// If we had a '[', read a multiplicity and ']'
	if (m) {
		if ( (*s) != ',' ) stop_error ("invalid arguments");
		write_next ();
		s = write_white ();
		read_expr (s,2);
		if ( (*s) !=']' ) stop_error ("invalid arguments");
		skip_next ();
		s = write_white ();
	}

	// Read a ":n" multiplier
	if ( (*s)==':' ) {
		skip_next ();
		s = write_white ();
		if (!m) write_string (",0");
		write_char (',');
		read_expr (s,2);
		s = write_white ();
	}
}



void read_expr (char *&s , int lvl)
{
	int num,l,i;
	char c;
	int ops = 0;
	int code;

	// Get codes for the beginning of division and exponentiation
	// expressions
	if (lvl==EXPLEVEL || lvl==DIVLEVEL) code = write_code ();

START:	ops++;
	if (lvl>MAXLEVEL) {
		// No binary/postfix operators have this level any more. Have
		// to read a term, prefix operator, reserved word, or bracket.
		if ( (l = read_op (s,num,PREFIX)) >= 0 ) {
			// Read a prefix operator, OK.
			write_stuff (s,num);
			read_expr (s,l);
			s = write_white ();
			return;
		}
		if ( read_op (s,num,RESERVED) >= 0 ) {
			// Read a reserved word, OK.
			write_stuff (s,num);
			return;
		}
		if ( read_op (s,num,GWTYPE) >= 0 ) {
			// Read a GW function, OK.
			write_stuff (s,num);
			c = read_open_bracket (s,num);
			if ( c!='(' ) stop_error ("invalid arguments");
			write_stuff (s,num);
			l = 0;
			while ( ! read_close_bracket (s,c,num) ) {
				if (l++) write_char ('+');
				write_string ("gwlist(");
				read_gwarg (s);
				write_char (')');
				s = write_white ();
				if ( read_close_bracket (s,c,num) ) break;
				if ( read_op(s,i,GWKEY)>=0 ) continue;
				if ( (*s)!=',' )
					stop_error ("invalid arguments");
				s = skip_next ();
				s = write_white ();
			}
			if (!l) write_string ("gwlist()");
			write_stuff (s,num); // closing bracket
			return;
		}
		if ( read_term (s,num) >= 0 ) {
			// Read name/number, OK.
			write_stuff (s,num);
			// A name may be followed by something in brackets.
			while ( c = read_open_bracket (s,num) ) {
				write_stuff (s,num); // write opening bracket
				while ( ! read_close_bracket (s,c,num) ) {
					read_expr (s,1); // read an expression
					s = write_white ();
				};
				write_stuff (s,num); // write closing bracket
			}
			return;
		}
		if ( c = read_open_bracket (s,num) ) {
			while ( c = read_open_bracket (s,num) ) {
				write_stuff (s,num); // write opening bracket
				while ( ! read_close_bracket (s,c,num) ) {
					read_expr (s,1); // read an expression
					s = write_white ();
				};
				write_stuff (s,num); // write closing bracket
			}
			return;
		}

		// Don't know what this is.
		stop_error ("parse error");
	}

	// Read a term of bigger level
	read_expr (s,lvl+1);
	if (lvl==EXPLEVEL && ops==2) write_char (')'); // last bracket for pow
	s = write_white ();

	// Read a binary or postfix operator of level lvl
	if ( lvl == read_op (s,num,POSTFIX) ) {
		write_stuff (s,num);
		return;
	}
	if ( lvl == read_op (s,num,BINARY) ) {
		if ( (*s) == '*' && num==1 ) {
			// The "*" operator is tricky, because it can be
			// prefix, postfix or binary. We have entries for
			// prefix and binary in the operator list. Postfix
			// is only possible if it is followed by a ')'. Check
			// this.
			write_stuff (s,num);
			if ( (*s) == ')' ) return; // postfix
			goto START; // binary
		}
		if (lvl==DIVLEVEL) {
			// Special division.
			if (num==2) { // "/." -> "/"
				write_next ();
				skip_next ();
				s = write_white ();
				goto START;
			}
			if (ops==1) { // "/" -> divconv...
				replace_code (code,"divconv(");
				write_char (')');
				write_stuff (s,num);
				goto START;
			}
			write_stuff (s,num);
			goto START;
		}
		if (lvl==EXPLEVEL) {
			// Special exponentiation "^" -> pow.
			if (ops!=1) stop_error ("invalid double exponents");
			replace_code (code,"pow(");
			write_char (',');
			skip_next ();
			s = write_white ();
			goto START;
		}
		write_stuff (s,num);
		goto START;
	}
}

