/*
 * inout.cc - functions for reading and writing files
 *
 */

#include <stdlib.h>
#include <iostream.h>
#include <assert.h>

#include "inout.h"



// Input and output pointers

static char *in_buf = NULL;
static int in_ptr = 0;
static int in_alloc = 0;
static char *out_buf = NULL;
static int out_ptr = 0;
static int out_alloc = 0;
static int line_nr = 0;

static int unused_codes = 0;



// Read an input line, terminated by newline or EOF

void in_line () {
	char c;
	char *s;
	int i;

	in_ptr = 0;
	line_nr++;

	if (cin.eof ()) {
		in_buf[0] = ENOF;
		return;
	}

	do {
		c = cin.get ();
		if (!cin) c=ENOF;
		in_buf[in_ptr++] = c;

		// Allocate more memory if necessary
		if (in_ptr>=in_alloc) {
			s = in_buf;
			in_alloc += 1024;
			assert (in_buf = new char[in_alloc]);
			for (i=0; i<in_ptr; i++) in_buf[i] = s[i];
			delete [] s;
		}

		if (cin.eof ()) break;
	} while (c!='\n' && c!=ENOF);

	in_ptr = 0;
	return;
}



// Initialization. Allocate some memory and return the first char, or NULL
// if there are no bytes in the input.

char *inout_init ()
{
	in_alloc = 1024;
	assert (in_buf = new char[1024]);
	out_alloc = 1024;
	assert (out_buf = new char[1024]);

	in_line();
	return in_buf;
}



void write_char (char c) // Write a character to the output queue
{
	char *s;
	int i;

	#ifdef INOUT_DEBUG
	if (c>=' ')
		cerr << "*inout* writing character " << c;
	else cerr << "*inout* writing control character " << (255&unsigned(c));
	cerr << " to the queue at position " << out_ptr << endl;
	#endif

	// Allocate more memory if necessary
	if (out_ptr>=out_alloc) {
		s = out_buf;
		out_alloc += 1024;
		assert (out_buf = new char[out_alloc]);
		for (i=0; i<out_ptr; i++) out_buf[i] = s[i];
		delete [] s;
	}

	out_buf[out_ptr++] = c;
}



void write_code (int n) // Write a (non-printable) code to the output queue
{
	int i;

	#ifdef INOUT_DEBUG
	cerr << "*inout* writing code " << n << " to the queue" << endl;
	#endif

	write_char (0);
	for (i=0; i<sizeof (int); i++) {
		write_char (char (n & 255));
		n>>=8;
	};
}



int write_code ()
{
	write_code (++unused_codes);
	return unused_codes;
}



void write_string (char *s) // write a constant string to the output buffer
{
	while (*s) {
		write_char (*s);
		s++;
	}
}



void write_flush () // flush the output buffer
{
	int i = 0;

	#ifdef INOUT_DEBUG
	cerr << "*inout* flushing output queue" << endl;
	#endif

	while (i<out_ptr) {
		if (out_buf[i]) cout << out_buf[i++];
		else i += sizeof (int) + 1; // ignore codes
	}

	out_ptr = 0; // reset queue
}



char *skip_next () // go to next input char, don't write current one to output
{
	if (in_buf[in_ptr]=='\n') in_line ();
	else in_ptr++;
	return &in_buf[in_ptr];
}



char *write_next () // go to next input char, write the current one
{
	write_char (in_buf[in_ptr]);
	return skip_next ();
}



char *write_white () // Copy whitespace
{
	unsigned char c;

	for (;;) {
		c = (unsigned char)(in_buf[in_ptr]);

		// Whitespace
		if ( c<=' ' ) {
			write_next ();
			continue;
		};

		// Backslash-Newline
		if ( c=='\\' && in_buf[in_ptr+1]=='\n' ) {
			write_next ();
			continue;
		}

		// "//" comment and '#' directive
		if ( c=='#' || (c=='/' && in_buf[in_ptr+1]=='/') ) {
			do {
				c = in_buf[in_ptr];
				if ((char)(c)==ENOF) break;
				write_next ();
				if ( c=='\\' && in_buf[in_ptr]=='\n' ) {
					write_next ();
					continue;
				}
			} while (in_buf[in_ptr]!='\n');
			continue;
		}

		// "/*" comment
		if ( c=='/' && in_buf[in_ptr+1]=='*' ) {
			write_next ();
			write_next ();
			for (;;) {
				c = in_buf[in_ptr];
				if ((char)(c)==ENOF) break;
				write_next ();
				if (c!='*') continue;
				if (in_buf[in_ptr]!='/') continue;
				write_next ();
				break;
			}
			continue;
		}

		break;
	}

	return &in_buf[in_ptr];
}



void replace_code (int n, char *s) // replace code with string
{
	char *t;
	int i,j,m;

	// Look for code in the output queue
	i = 0;
	do {
		if (i>=out_ptr) {
			#ifdef INOUT_DEBUG
			cout << "error: cannot find control code " << n
			     << endl;
			#endif
			stop_error ("internal error"); // not found
		}
		if (out_buf[i++]) continue;

		// Read code
		m = 0;
		for (j=0; j<sizeof(int); j++)
			m |= ((unsigned char)(out_buf[i++]) << (8*j));
	} while (m != n);

	// Code found.
	i -= (sizeof(int)+1); // the code starts at out_buf[i];
	j = strlen (s)-sizeof(int)-1; // the queue has to be moved
				      // by this many chars for the string
				      // to fit.

	if (out_ptr+j>=out_alloc) {
		// we have to allocate more memory for this.
		t = out_buf;
		out_alloc = out_ptr+j+1024;
		assert (out_buf = new char[out_alloc]);
		for (m=0; m<out_ptr; m++) out_buf[m] = t[m];
		delete [] t;
	}

	// Now move the queue
	if (j>0) {
		// move forward
		for (m=out_ptr-1; m>=i ; m--) out_buf[m+j] = out_buf[m];
	} else if (j<0) {
		// move back
		for (m=i-j; m<out_ptr ; m++) out_buf[m+j] = out_buf[m];
	}
	out_ptr += j;

	// Write string
	for (m=0; m<strlen(s); m++) out_buf[i+m] = s[m];

	#ifdef INOUT_DEBUG
	cerr << "*inout* code " << n << " replaced with " << s << endl;
	#endif
}



void stop_error (char *s)
{
	int i;

	cerr << "*** " << s << " in line " << line_nr << ":" << endl;

	i = 0;
	while (in_buf[i]!=ENOF && in_buf[i] !='\n') cerr << in_buf[i++];
	cerr << endl;

	i = 0;
	while (in_buf[i]!=ENOF && in_buf[i]!='\n') {
		if (i == in_ptr) cerr << "^";
		else if (in_buf[i] == '\t') cerr << in_buf[i];
		else cerr << " ";
		i++;
	}
	if ((i++) == in_ptr) cerr << "^";
	cerr << endl;

	exit (1);
}

